@@ -1,20 +1,18 @@
-use crate::{PreKeyBundle, Signature, X3DHError, X3dhVault, CSUITE, ENROLLMENT_MSG_SIZE};
+use crate::{PreKeyBundle, Signature, X3DHError, X3dhVault, CSUITE};
 use arrayref::array_ref;
+use ockam_core::Result;
 use ockam_key_exchange_core::{CompletedKeyExchange, KeyExchanger};
 use ockam_vault_core::{
     PublicKey, Secret, SecretAttributes, SecretPersistence, SecretType, AES256_SECRET_LENGTH,
     CURVE25519_SECRET_LENGTH,
 };
-use subtle::ConstantTimeEq;
 
 #[derive(Debug)]
 enum ResponderState {
-    /// Create a PreKey Bundle
-    GenerateBundle,
     /// Expect an enrollment message from this EIK
-    SetEnrollmentKey,
-    /// Verify an enrollment message
-    VerifyEnrollment,
+    HandleInitiatorKeys,
+    /// Create a PreKey Bundle
+    SendBundle,
     /// Done
     Done,
 }
@@ -22,12 +20,9 @@ enum ResponderState {
 /// The responder of X3DH creates a prekey bundle that can be used to establish a shared
 /// secret key with another party that can use
 pub struct Responder<V: X3dhVault> {
-    // Identity key and signer prekey are wrapped in Arc because they are possible shared
-    // among threads/modules
     identity_key: Option<Secret>,
     signed_prekey: Option<Secret>,
     one_time_prekey: Option<Secret>,
-    expected_enrollment_key: Option<PublicKey>,
     state: ResponderState,
     vault: V,
     completed_key_exchange: Option<CompletedKeyExchange>,
@@ -39,14 +34,13 @@ impl<V: X3dhVault> Responder<V> {
             identity_key,
             signed_prekey: None,
             one_time_prekey: None,
-            expected_enrollment_key: None,
             completed_key_exchange: None,
-            state: ResponderState::GenerateBundle,
+            state: ResponderState::HandleInitiatorKeys,
             vault,
         }
     }
 
-    fn prologue(&mut self) -> ockam_core::Result<()> {
+    fn prologue(&mut self) -> Result<()> {
         let p_atts = SecretAttributes::new(
             SecretType::Curve25519,
             SecretPersistence::Persistent,
@@ -62,8 +56,6 @@ impl<V: X3dhVault> Responder<V> {
         }
         self.signed_prekey = Some(self.vault.secret_generate(p_atts)?);
         self.one_time_prekey = Some(self.vault.secret_generate(e_atts)?);
-        self.expected_enrollment_key = None;
-        self.completed_key_exchange = None;
         Ok(())
     }
 }
@@ -75,14 +67,12 @@ impl<V: X3dhVault> std::fmt::Debug for Responder<V> {
             r#"X3dhResponder {{ identity_key: {:?},
                                       signed_prekey: {:?},
                                       one_time_prekey: {:?},
-                                      expected_enrollment_key: {:?},
                                       state: {:?},
                                       vault,
                                       completed_key_exchange: {:?} }}"#,
             self.identity_key,
             self.signed_prekey,
             self.one_time_prekey,
-            self.expected_enrollment_key,
             self.state,
             self.completed_key_exchange
         )
@@ -90,10 +80,9 @@ impl<V: X3dhVault> std::fmt::Debug for Responder<V> {
 }
 
 impl<V: X3dhVault> KeyExchanger for Responder<V> {
-    fn process(&mut self, data: &[u8]) -> ockam_core::Result<Vec<u8>> {
+    fn generate_request(&mut self, _payload: &[u8]) -> Result<Vec<u8>> {
         match self.state {
-            ResponderState::GenerateBundle => {
-                self.prologue()?;
+            ResponderState::SendBundle => {
                 let identity_secret_key =
                     self.identity_key.as_ref().ok_or(X3DHError::InvalidState)?;
                 let signed_prekey = self.signed_prekey.as_ref().ok_or(X3DHError::InvalidState)?;
@@ -113,44 +102,48 @@ impl<V: X3dhVault> KeyExchanger for Responder<V> {
                     signature_prekey: Signature(signature),
                     one_time_prekey: one_time_prekey_pub,
                 };
-                self.state = ResponderState::SetEnrollmentKey;
+                self.state = ResponderState::Done;
                 Ok(bundle.to_bytes())
             }
-            ResponderState::SetEnrollmentKey => {
-                if data.len() != 32 {
-                    return Err(X3DHError::MessageLenMismatch.into());
-                }
-                self.expected_enrollment_key =
-                    Some(PublicKey::new(array_ref![data, 0, 32].to_vec()));
-                self.state = ResponderState::VerifyEnrollment;
-                Ok(vec![])
+            ResponderState::HandleInitiatorKeys | ResponderState::Done => {
+                Err(X3DHError::InvalidState.into())
             }
-            ResponderState::VerifyEnrollment => {
-                if data.len() != ENROLLMENT_MSG_SIZE {
+        }
+    }
+
+    fn handle_response(&mut self, response: &[u8]) -> Result<Vec<u8>> {
+        match self.state {
+            ResponderState::HandleInitiatorKeys => {
+                if response.len() != 64 {
                     return Err(X3DHError::MessageLenMismatch.into());
                 }
+                self.prologue()?;
+
+                let other_identity_pubkey = PublicKey::new(array_ref![response, 0, 32].to_vec());
+                let other_ephemeral_pubkey = PublicKey::new(array_ref![response, 32, 32].to_vec());
+
                 let signed_prekey = self.signed_prekey.as_ref().ok_or(X3DHError::InvalidState)?;
                 let one_time_prekey = self
                     .one_time_prekey
                     .as_ref()
                     .ok_or(X3DHError::InvalidState)?;
-                let eik = self
-                    .expected_enrollment_key
-                    .as_ref()
-                    .ok_or(X3DHError::InvalidState)?;
-                let id = self.vault.sha256(eik.as_ref())?;
-                if id.ct_eq(&data[32..64]).unwrap_u8() != 1 {
-                    return Err(X3DHError::InvalidHash.into());
-                }
-                let ek = PublicKey::new(array_ref![data, 0, 32].to_vec());
-                let local_static_secret =
-                    self.identity_key.take().ok_or(X3DHError::InvalidState)?;
 
-                let dh1 = self.vault.ec_diffie_hellman(signed_prekey, &eik)?;
-                let dh2 = self.vault.ec_diffie_hellman(&local_static_secret, &ek)?;
-                let dh3 = self.vault.ec_diffie_hellman(signed_prekey, &ek)?;
-                let dh4 = self.vault.ec_diffie_hellman(one_time_prekey, &ek)?;
-                let mut ikm_bytes = vec![0xFFu8; 32];
+                let local_static_secret =
+                    self.identity_key.as_ref().ok_or(X3DHError::InvalidState)?;
+
+                let dh1 = self
+                    .vault
+                    .ec_diffie_hellman(signed_prekey, &other_identity_pubkey)?;
+                let dh2 = self
+                    .vault
+                    .ec_diffie_hellman(local_static_secret, &other_ephemeral_pubkey)?;
+                let dh3 = self
+                    .vault
+                    .ec_diffie_hellman(signed_prekey, &other_ephemeral_pubkey)?;
+                let dh4 = self
+                    .vault
+                    .ec_diffie_hellman(one_time_prekey, &other_ephemeral_pubkey)?;
+                let mut ikm_bytes = vec![0xFFu8; 32]; // FIXME
                 ikm_bytes.extend_from_slice(self.vault.secret_export(&dh1)?.as_ref());
                 ikm_bytes.extend_from_slice(self.vault.secret_export(&dh2)?.as_ref());
                 ikm_bytes.extend_from_slice(self.vault.secret_export(&dh3)?.as_ref());
@@ -183,31 +176,17 @@ impl<V: X3dhVault> KeyExchanger for Responder<V> {
                 state_hash.append(&mut ikm_bytes);
                 let state_hash = self.vault.sha256(state_hash.as_slice())?;
 
-                let mut aad = data[..64].to_vec();
-                aad.extend_from_slice(CSUITE);
-                aad.extend_from_slice(&state_hash);
-                //TODO: get the channel address from the message somehow if needed
-                let plaintext = self.vault.aead_aes_gcm_decrypt(
-                    &decrypt_key,
-                    &data[64..],
-                    &data[..12],
-                    aad.as_slice(),
-                )?;
-                let ikb = PublicKey::new(array_ref![plaintext, 0, 32].to_vec());
-                let signature = array_ref![plaintext, 32, 64];
-                self.vault.verify(signature, &eik, &plaintext[..32])?;
-
                 self.completed_key_exchange = Some(CompletedKeyExchange::new(
                     state_hash,
                     encrypt_key,
                     decrypt_key,
-                    local_static_secret,
-                    ikb,
                 ));
-                self.state = ResponderState::Done;
+                self.state = ResponderState::SendBundle;
                 Ok(vec![])
             }
-            ResponderState::Done => Err(X3DHError::InvalidState.into()),
+            ResponderState::SendBundle | ResponderState::Done => {
+                Err(X3DHError::InvalidState.into())
+            }
         }
     }
 
@@ -215,7 +194,7 @@ impl<V: X3dhVault> KeyExchanger for Responder<V> {
         matches!(self.state, ResponderState::Done)
     }
 
-    fn finalize(self) -> ockam_core::Result<CompletedKeyExchange> {
+    fn finalize(self) -> Result<CompletedKeyExchange> {
         self.completed_key_exchange
             .ok_or_else(|| X3DHError::InvalidState.into())
     }
